@@ -7,6 +7,8 @@ import numpy as np
 from numpy import sin, cos
 import matplotlib.pyplot as plt
 from itertools import combinations
+from numba import njit
+import math
 
 from .spatial_graph_diagrams import (Vertex, Crossing, SpatialGraphDiagram)
 
@@ -23,6 +25,9 @@ class AbstractGraph:
         self.nodes = self._validate_nodes(nodes)
         self.edges = self._validate_edges(edges)
 
+        self.adjacent_edge_pairs = self._get_adjacent_edge_pairs()
+        self.nonadjacent_edge_pairs = [edge_pair for edge_pair in self.edge_pairs if edge_pair not in self.adjacent_edge_pairs]
+
     @staticmethod
     def _validate_nodes(nodes: list[str]) -> list[str]:
         """
@@ -113,8 +118,7 @@ class AbstractGraph:
     def edge_pairs(self):
         return list(combinations(self.edges, 2))
 
-    @property
-    def adjacent_edge_pairs(self):
+    def _get_adjacent_edge_pairs(self):
 
         adjacent_edge_pairs = []
 
@@ -162,6 +166,7 @@ class LinearAlgebra:
         return next(self.rotation_generator_object)
 
     @staticmethod
+    # @njit(cache=True)
     def rotate(positions: np.ndarray,
                rotation:  np.ndarray) -> np.ndarray:
         """
@@ -206,7 +211,9 @@ class LinearAlgebra:
 
         return rotated_node_positions
 
+
     @staticmethod
+    @njit(cache=True)
     def get_line_segment_intersection(a: np.ndarray,
                                       b: np.ndarray,
                                       c: np.ndarray,
@@ -270,7 +277,7 @@ class LinearAlgebra:
         S1  = np.dot(d1, d12.T)
         S2  = np.dot(d2, d12.T)
         R   = np.dot(d1, d2.T)
-        den = np.dot(D1, D2) - np.square(R)
+        den = D1 * D2 - R**2
 
         # Check if one or both line segments are points
         if D1 == 0. or D2 == 0.:
@@ -321,15 +328,14 @@ class LinearAlgebra:
 
         min_dist = np.linalg.norm(d1 * t - d2 * u - d12)
 
-        if not np.isclose(min_dist, 0):
-            min_dist_position = None
-        else:
-            min_dist_position = np.array([a + d1 * t]).reshape(-1)
+        # min_dist_position = np.array([a + d1 * t]).reshape(-1)
+        min_dist_position = a + d1 * t
 
         return min_dist, min_dist_position
 
 
     @staticmethod
+    # @njit(cache=True)
     def calculate_intermediate_y_position(a:     np.ndarray,
                                           b:     np.ndarray,
                                           x_int: float,
@@ -343,12 +349,11 @@ class LinearAlgebra:
 
         delta_x = x2 - x1
         delta_y = y2 - y1
-        delta_z = z2 - z1
+        # delta_z = z2 - z1
 
         ratio_x = (x_int - x1) / delta_x
-        ratio_z = (z_int - z1) / delta_z
-
-        assert np.isclose(ratio_x, ratio_z, rtol=1e-3)
+        # ratio_z = (z_int - z1) / delta_z
+        # assert np.isclose(ratio_x, ratio_z, rtol=1e-3)
 
         y_int = y1 + ratio_x * delta_y
 
@@ -360,6 +365,7 @@ class LinearAlgebra:
 
 
     @staticmethod
+    # @njit(cache=True)
     def calculate_counter_clockwise_angle(vector_a: np.ndarray,
                                           vector_b: np.ndarray) -> float:
         """
@@ -810,17 +816,21 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
 
         return overlap_order
 
+
+    # @njit
     def get_crossings(self):
 
         crossing_num = 0
-        crossings = []
+        crossings    = []
         crossing_edge_pairs = []
         crossing_positions = []
 
-        nonadjacent_edge_pairs = [edge_pair for edge_pair in self.edge_pairs if
-                                  edge_pair not in self.adjacent_edge_pairs]
+        # TODO Modify nonadjacent edge pairs that are within some axis aligned bounding box
 
-        for line_1, line_2 in nonadjacent_edge_pairs:
+
+
+
+        for line_1, line_2 in self.nonadjacent_edge_pairs:
 
             a = self.projected_node_positions[self.nodes.index(line_1[0])]
             b = self.projected_node_positions[self.nodes.index(line_1[1])]
@@ -829,23 +839,25 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
 
             min_dist, crossing_position = self.get_line_segment_intersection(a, b, c, d)
 
+            if min_dist > 0.0001:
+                crossing_position = None
+
             if crossing_position is np.inf:
                 raise ValueError('The edges are overlapping. This is not a valid spatial graph.')
 
             elif crossing_position is None:
                 pass
 
-            elif type(crossing_position) is np.ndarray and np.isclose(min_dist, 0):
+            elif type(crossing_position) is np.ndarray and min_dist < 0.0001:
                 crossings.append('crossing_' + str(crossing_num))
                 crossing_num += 1
                 crossing_positions.append(crossing_position)
                 crossing_edge_pair = self.get_crossing_edge_order(line_1, line_2, crossing_position)
                 crossing_edge_pairs.append(crossing_edge_pair)
 
-
         return crossings, crossing_positions, crossing_edge_pairs
 
-    def project(self):
+    def project(self, max_iter=50, predefined_rotation=None):
         """
         Project the spatial graph onto a random 2D plane.
 
@@ -861,12 +873,17 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
 
         """
 
-        # Initialize while loop exit conditions
-        valid_projection = False
-        iter = 0
-        max_iter = 10
+        if predefined_rotation is not None:
+            self.rotation = predefined_rotation
+            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
+            self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()
+            return
 
-        while not valid_projection and iter < max_iter:
+
+        rotations = []
+        num_crossings = []
+
+        for _ in range(max_iter):
 
             try:
 
@@ -892,11 +909,14 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
                     d = self.projected_node_positions[self.nodes.index(line_2[1])]
                     min_dist, crossing_position = self.get_line_segment_intersection(a, b, c, d)
 
+                    if min_dist > 0.0001:
+                        crossing_position = None
+
                     if crossing_position is not None and crossing_position is not np.inf:
-                        assertion_1 = all(np.isclose(crossing_position, a))
-                        assertion_2 = all(np.isclose(crossing_position, b))
-                        assertion_3 = all(np.isclose(crossing_position, c))
-                        assertion_4 = all(np.isclose(crossing_position, d))
+                        assertion_1 = all((crossing_position - a) < 0.0001)
+                        assertion_2 = all((crossing_position - b) < 0.0001)
+                        assertion_3 = all((crossing_position - c) < 0.0001)
+                        assertion_4 = all((crossing_position - d) < 0.0001)
                         if not any([assertion_1, assertion_2, assertion_3, assertion_4]):
                             raise ValueError('Adjacent edges must intersect at the endpoints.')
 
@@ -912,17 +932,28 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
 
                 self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()
 
-                # If we made it this far without raising an exception, then we have a valid projection
-                valid_projection = True
+                rotations.append(self.rotation)
+                num_crossings.append(len(self.crossings))
 
             except ValueError:
                 self.rotation = self.random_rotation()
                 self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
 
-                iter += 1
+            # If there is no error in the projection rotate anyways
+            self.rotation = self.random_rotation()
+            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
 
-        if iter == max_iter:
+
+
+        if len(rotations) == 0:
             raise Exception('Could not find a valid rotation after {} iterations'.format(max_iter))
+        else:
+            min_crossings = min(num_crossings)
+            min_crossings_index = num_crossings.index(min_crossings)
+            self.rotation = rotations[min_crossings_index]
+            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
+            self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()
+
 
     def create_vertices(self):
         """
@@ -994,10 +1025,10 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
         """
 
         fig = plt.figure()
-        ax2 = plt.subplot()
+        # ax2 = plt.subplot()
 
-        # ax1 = plt.subplot(221, projection='3d')
-        # ax2 = plt.subplot(222)
+        ax1 = plt.subplot(221, projection='3d')
+        ax2 = plt.subplot(222)
         # ax3 = plt.subplot(223, projection='3d')
         #
         # # Axis 1
@@ -1040,11 +1071,11 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
         # # Figure layout
         # # plt.tight_layout(pad=2, w_pad=7, h_pad=0)
         #
-        # # Plot 3D
-        # for edge in self.edges:
-        #     point_1 = self.node_positions[self.nodes.index(edge[0])]
-        #     point_2 = self.node_positions[self.nodes.index(edge[1])]
-        #     ax1.plot3D([point_1[0], point_2[0]], [point_1[1], point_2[1]], [point_1[2], point_2[2]])
+        # Plot 3D
+        for edge in self.edges:
+            point_1 = self.node_positions[self.nodes.index(edge[0])]
+            point_2 = self.node_positions[self.nodes.index(edge[1])]
+            ax1.plot3D([point_1[0], point_2[0]], [point_1[1], point_2[1]], [point_1[2], point_2[2]])
         #
         # # Shrink current axis by 40%
         # box = ax1.get_position()
@@ -1081,7 +1112,7 @@ class SpatialGraph(AbstractGraph, LinearAlgebra):
 
         # Put a legend to the right of the current axis
         # ax2.legend(self.edges, loc='center left', bbox_to_anchor=(1, -0.25))
-        ax2.legend(self.edges, loc='center left', bbox_to_anchor=(0.75, 0.5), fontsize=9)
+        # ax2.legend(self.edges, loc='center left', bbox_to_anchor=(0.75, 0.5), fontsize=9)
         # ax2.legend(self.edges)
 
 
